from pathlib import Path
from .suite import Suite
from .utils import *
from .pipeline import *
from typing import Tuple
from alive_progress import alive_bar


class Workspace:
    def __init__(self, dir: str):
        self.dir = Path(dir).absolute()
        # make sure the workspace directory exists.
        self.dir.mkdir(exist_ok=True)

        self.src_dir = self.dir / 'src'
        self.src_dir.mkdir(exist_ok=True)

        self.bin_dir = self.dir / 'bin'
        self.bin_dir.mkdir(exist_ok=True)

        self.ir_dir = self.dir / 'ir'
        self.ir_dir.mkdir(exist_ok=True)

        self.suites = []

    def add_suite(self, suite, *args, **kwargs):
        s = suite(self)
        s.configure(*args, **kwargs)
        self.suites.append(s)

    def dump_benchmarks(self):
        for suite in self.suites:
            print(f'suite {suite.name}')
            for benchmark in suite.benchmarks:
                print(f'  {benchmark.name}')

    def extract_bitcode(self, input, output):
        self.shell('get-bc', '-o', output, input)
        self.shell('llvm-dis', input)

    def run_jobs(self, jobs):
        if len(jobs) > 0:
            with alive_bar(len(jobs)) as bar:  # declare your expected total
                for jobname, op, *args in jobs:
                    bar.text(jobname)
                    op(*args)
                    bar()

    def prepare(self):
        # a list of jobs to work on. This will be appended to and
        # worked through before a pipeline can run.
        jobs: Tuple[str, function] = []

        # first, make sure all the benchmarks are acquired. We just do this by
        # checking if the directory of the suite exists.
        for suite in self.suites:
            if not suite.src.exists():
                jobs.append((f'acquire {suite.name}', suite.acquire))

        # second, compile the a.out files for each benchmark suite. If the a.out
        # file doesn't exist, add it to the job list.
        for suite in self.suites:
            suite_bin = self.bin_dir / suite.name
            suite_bin.mkdir(exist_ok=True)
            jobs += suite.get_compile_jobs()

        # Third, make sure input.bc exists in each ir/<suite>/<benchmark>/ folder
        for suite in self.suites:
            suite_ir = self.ir_dir / suite.name
            suite_ir.mkdir(exist_ok=True)
            suite_bin = self.bin_dir / suite.name
            suite_bin.mkdir(exist_ok=True)
            for benchmark in suite.benchmarks:
                benchmark_ir_dir = suite_ir / benchmark.name
                benchmark_ir_dir.mkdir(exist_ok=True)

                input = suite_bin / benchmark.name / 'a.out'
                output = benchmark_ir_dir / 'input.bc'
                if not output.exists():
                    jobs.append(
                        (f'extract {suite.name}/{benchmark.name}', self.extract_bitcode, input, output))

        # Execute those jobs we just made.
        self.run_jobs(jobs)

    def run_pipeline(self, pipeline: Pipeline):
        # Make sure everything is setup!
        self.prepare()
        # Now run the pipeline on every benchmark's IR
        jobs = []
        for suite in self.suites:
            suite_ir = self.ir_dir / suite.name
            for benchmark in suite.benchmarks:
                benchmark_ir_dir = suite_ir / benchmark.name
                benchmark_ir_input = benchmark_ir_dir / 'input.bc'

                output = benchmark_ir_dir / f'{pipeline.name}.bc'
                jobs += pipeline.create_jobs(benchmark_ir_input,
                                             output, benchmark)
                benchmark_link_dest = self.bin_dir / suite.name / benchmark.name / pipeline.name
                jobs.append((f'link {suite.name}/{benchmark.name} for pipeline {pipeline.name}',
                            benchmark.link, output, benchmark_link_dest))

        self.run_jobs(jobs)

    def shell(self, *args):
        # print('running: ', *args)
        with open(self.dir / 'output.txt', 'a+') as out:
            out.write(f'XXXXXX RUNNING COMMAND {args}\n')
            proc = subprocess.Popen(args, stdout=out, stderr=out)
            proc.wait()
